# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from __future__ import division
import threading
from pai4sk.ensemble.forest import ForestClassifier
import numpy as np
from scipy.sparse import csr_matrix 
from .utils import _check_snapml_types, _check_snapml_type, _is_gpu_matrix_type, _get_data_size, _check_uparams
from _snap_ml.utils import _is_mpi_enabled, _is_data_snapml_partition_type
from ...utils import check_random_state, check_array, compute_sample_weight
from ...exceptions import NotFittedError
from ...tree import DecisionTreeClassifier
import time
import warnings
import numbers

_mpi_enabled = _is_mpi_enabled()

class RandomForestClassifier(ForestClassifier):
    """A random forest classifier.

    A random forest is a meta estimator that fits a number of decision tree
    classifiers on various sub-samples of the dataset and uses averaging to
    improve the predictive accuracy and control over-fitting.
    The sub-sample size is always the same as the original
    input sample size but the samples are drawn with replacement if
    `bootstrap=True` (default).

    Read more in the :ref:`User Guide <forest>`.

    Parameters
    ----------
    n_estimators : integer, optional (default=10)
        The number of trees in the forest.

        .. versionchanged:: 0.20
           The default value of ``n_estimators`` will change from 10 in
           version 0.20 to 100 in version 0.22.

    criterion : string, optional (default="gini")
        The function to measure the quality of a split. Supported criteria are
        "gini" for the Gini impurity and "entropy" for the information gain.
        Note: this parameter is tree-specific.

    max_depth : integer or None, optional (default=None)
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int, float, optional (default=2)
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_samples_leaf : int, float, optional (default=1)
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_weight_fraction_leaf : float, optional (default=0.)
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : int, float, string or None, optional (default="auto")
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `int(max_features * n_features)` features are considered at each
          split.
        - If "auto", then `max_features=sqrt(n_features)`.
        - If "sqrt", then `max_features=sqrt(n_features)` (same as "auto").
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_leaf_nodes : int or None, optional (default=None)
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, optional (default=0.)
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

        .. versionadded:: 0.19

    min_impurity_split : float, (default=1e-7)
        Threshold for early stopping in tree growth. A node will split
        if its impurity is above the threshold, otherwise it is a leaf.

        .. deprecated:: 0.19
           ``min_impurity_split`` has been deprecated in favor of
           ``min_impurity_decrease`` in 0.19. The default value of
           ``min_impurity_split`` will change from 1e-7 to 0 in 0.23 and it
           will be removed in 0.25. Use ``min_impurity_decrease`` instead.


    bootstrap : boolean, optional (default=True)
        Whether bootstrap samples are used when building trees. If False, the
        whole datset is used to build each tree.

    oob_score : bool (default=False)
        Whether to use out-of-bag samples to estimate
        the generalization accuracy.

    n_jobs : int or None, optional (default=None)
        The number of jobs to run in parallel for both `fit` and `predict`.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    verbose : int, optional (default=0)
        Controls the verbosity when fitting and predicting.

    warm_start : bool, optional (default=False)
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit a whole
        new forest. See :term:`the Glossary <warm_start>`.

    class_weight : dict, list of dicts, "balanced", "balanced_subsample" or \
    None, optional (default=None)
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have weight one. For
        multi-output problems, a list of dicts can be provided in the same
        order as the columns of y.

        Note that for multioutput (including multilabel) weights should be
        defined for each class of every column in its own dict. For example,
        for four-class multilabel classification weights should be
        [{0: 1, 1: 1}, {0: 1, 1: 5}, {0: 1, 1: 1}, {0: 1, 1: 1}] instead of
        [{1:1}, {2:5}, {3:1}, {4:1}].

        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``

        The "balanced_subsample" mode is the same as "balanced" except that
        weights are computed based on the bootstrap sample for every tree
        grown.

        For multi-output, the weights of each column of y will be multiplied.

        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.

    use_gpu : boolean, default : False
        Flag that indicates the hardware platform used for training. If True, the training
        is performed using the GPU. If False, the training is performed using the CPU. Currently
        only CPU training is supported.

    hist_type: string, or None, optional, default : None
        Indicates whether to use histograms or not. Default is None and does not use histograms.
        - If 'Auto' or 'Quantiles' it will use histograms with Quantiles

    hist_nbins: int, or None, optional, default : None
        Indicates the number of bins to be used in histogram mode.
        Only valid if hist_type is not None. Default value if hist_type is not None is 64.


    Attributes
    ----------
    estimators_ : list of DecisionTreeClassifier
        The collection of fitted sub-estimators.

    classes_ : array of shape = [n_classes] or a list of such arrays
        The classes labels (single output problem), or a list of arrays of
        class labels (multi-output problem).

    n_classes_ : int or list
        The number of classes (single output problem), or a list containing the
        number of classes for each output (multi-output problem).

    n_features_ : int
        The number of features when ``fit`` is performed.

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    feature_importances_ : array of shape = [n_features]
        The feature importances (the higher, the more important the feature).

    oob_score_ : float
        Score of the training dataset obtained using an out-of-bag estimate.

    oob_decision_function_ : array of shape = [n_samples, n_classes]
        Decision function computed with out-of-bag estimate on the training
        set. If n_estimators is small it might be possible that a data point
        was never left out during the bootstrap. In this case,
        `oob_decision_function_` might contain NaN.

    Examples
    --------
    >>> from pai4sk.ensemble import RandomForestClassifier
    >>> from pai4sk.datasets import make_classification

    >>> X, y = make_classification(n_samples=1000, n_features=4,
    ...                            n_informative=2, n_redundant=0,
    ...                            random_state=0, shuffle=False)
    >>> clf = RandomForestClassifier(n_estimators=100, max_depth=2,
    ...                              random_state=0)
    >>> clf.fit(X, y)
    RandomForestClassifier(bootstrap=True, class_weight=None, criterion='gini',
                max_depth=2, max_features='auto', max_leaf_nodes=None,
                min_impurity_decrease=0.0, min_impurity_split=None,
                min_samples_leaf=1, min_samples_split=2,
                min_weight_fraction_leaf=0.0, n_estimators=100, n_jobs=None,
                oob_score=False, random_state=0, verbose=0, warm_start=False)
    >>> print(clf.feature_importances_)
    [0.14205973 0.76664038 0.0282433  0.06305659]
    >>> print(clf.predict([[0, 0, 0, 0]]))
    [1]

    Notes
    -----
    The default values for the parameters controlling the size of the trees
    (e.g. ``max_depth``, ``min_samples_leaf``, etc.) lead to fully grown and
    unpruned trees which can potentially be very large on some data sets. To
    reduce memory consumption, the complexity and size of the trees should be
    controlled by setting those parameter values.

    The features are always randomly permuted at each split. Therefore,
    the best found split may vary, even with the same training data,
    ``max_features=n_features`` and ``bootstrap=False``, if the improvement
    of the criterion is identical for several splits enumerated during the
    search of the best split. To obtain a deterministic behaviour during
    fitting, ``random_state`` has to be fixed.

    References
    ----------

    .. [1] L. Breiman, "Random Forests", Machine Learning, 45(1), 5-32, 2001.

    See also
    --------
    DecisionTreeClassifier, ExtraTreesClassifier
    """

    #__slots__ = ('_randomforestsklearn')

    _param_list = []    # Store original arguments
    _uparams = None     # Store user provided arguments

    # Override the existing function to include the current list of parameters
    # as the init signature is modified by the decorator
    def _get_param_names(self):
        return self._param_list


    @_check_uparams
    def __init__(self,
                 n_estimators='warn',
                 criterion="gini",                   
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 min_impurity_split=None,
                 bootstrap=True,
                 oob_score=False,
                 n_jobs=None,
                 random_state=None,
                 verbose=0,
                 warm_start=False,
                 class_weight=None,
                 algo='default',
                 use_gpu=False,                 #pai4sk specific
                 hist_type=None,                #pai4sk specific
                 hist_nbins=None):              #pai4sk specific
        super(RandomForestClassifier, self).__init__(
           base_estimator=DecisionTreeClassifier(),
           n_estimators=n_estimators,
           estimator_params=("criterion", "max_depth", "min_samples_split",
                              "min_samples_leaf", "min_weight_fraction_leaf",
                              "max_features", "max_leaf_nodes",
                              "min_impurity_decrease", "min_impurity_split",
                              "random_state"),
           bootstrap=bootstrap,
           oob_score=oob_score,
           n_jobs=n_jobs,
           random_state=random_state,
           verbose=verbose,
           warm_start=warm_start,
           class_weight=class_weight)

        self.n_estimators = n_estimators
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.min_impurity_split = min_impurity_split
        self.oob_score = oob_score
        self.warm_start = warm_start
        self.class_weight = class_weight
        self.use_gpu = use_gpu
        self.verbose = verbose
        self.bootstrap = bootstrap
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.hist_type = hist_type
        self.hist_nbins = hist_nbins
        self.algo = algo 
        self._snaprf = None
        self._randomforestsklearn = super(RandomForestClassifier, self)

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._randomforestsklearn = super(RandomForestClassifier, self)


    @property
    def feature_importances_(self):

        if self._snaprf:
           raise ValueError("feature_importances_ is not supported in snap-ml ; use RandomForestClassifierSklearn otherwise")
           return np.zeros(self.n_features_, dtype=np.float64)
        else:
           return ForestClassifier.feature_importances_.fget(self)

           
    def _check_set_snapml(self, X, y, sample_weight=None):
        
        snap_args = {}
        mod_params = []
        unsupported_params = []

        if (self.algo == 'default'):
            return False

        if (not self.min_weight_fraction_leaf == 0.) or 'min_weight_fraction_leaf' in self._uparams:
            unsupported_params.append("min_weight_fraction_leaf")

        if (not self.max_leaf_nodes is None) or 'max_leaf_nodes' in self._uparams:
            unsupported_params.append("max_leaf_nodes")

        if (not self.min_impurity_decrease == 0.) or 'min_impurity_decrease' in self._uparams:
            unsupported_params.append("min_impurity_decrease")

        if (not self.min_impurity_split is None) or 'min_impurity_split' in self._uparams:
            unsupported_params.append("min_impurity_split")

        if (not self.oob_score is False) or 'oob_score' in self._uparams:
            unsupported_params.append("oob_score")

        if  (not self.warm_start is False) or 'warm_start' in self._uparams:
            unsupported_params.append("warm_start")

        if (not self.class_weight is None) or 'class_weight' in self._uparams: 
            unsupported_params.append("class_weight")

        if 'min_samples_split' in self._uparams:
            unsupported_params.append("min_samples_split")

        if (not self.algo == 'snapml'):
            unsupported_params.append("algo")

        if (not self.max_features in [None, 'log2', 'sqrt', 'auto']):
            if (not isinstance(self.max_features, (int, float))):
                unsupported_params.append("max_features")

        if (not self.criterion == 'gini'):
            unsupported_params.append("criterion is not gini")

        if not _check_snapml_types(X,y):
            unsupported_params.append("X/y")
        else: 
            uniq_labs = np.unique(y)
            if (len(uniq_labs) == 1 ):
               unsupported_params.append("unsupported y:only one label")
            if (len(uniq_labs) > 2 ):
               unsupported_params.append("y:Multiclass classification not supported")

        if  _is_data_snapml_partition_type(X) or _is_gpu_matrix_type(X, y):
            unsupported_params.append("X/y")

        #type check for pai4sk - to give user a fallback chance
        if (not self.max_depth is None) and (not isinstance(self.max_depth, (int))):
            unsupported_params.append("max_depth")


        # pradghos: Place Holder case for Future implementation.
        if (self.use_gpu is True):
            raise ValueError("pai4sk: currently RandomForestClassifier supports only CPU")

        #min_samples_leaf validation
        if (not isinstance(self.min_samples_leaf, (int))) and (not isinstance(self.min_samples_leaf, (float))):
            unsupported_params.append("min_samples_leaf:invalid type")
        if (isinstance(self.min_samples_leaf, (int))) and (self.min_samples_leaf < 1):
            unsupported_params.append("min_samples_leaf:invalid type")
        if (isinstance(self.min_samples_leaf, (float))) and ((self.min_samples_leaf <= 0) or (self.min_samples_leaf > 1)):
            unsupported_params.append("min_samples_leaf:invalid type")
 

        if not unsupported_params:

            # n_estimator value does match with snap-ml default value.
            # hence it has been modified if user does not have any explicit 
            # input.
            if 'n_estimators' in self._uparams:
                snap_args['n_estimators'] = self.n_estimators
            else:
                mod_params.append("n_estimators")
                snap_args['n_estimators'] = 10


            #snap_args 
            snap_args['verbose'] = False if self.verbose == 0 else True    
            snap_args['criterion'] = self.criterion                      # snap ml default value                                                     
            snap_args['max_depth'] = self.max_depth
            snap_args['min_samples_leaf'] = self.min_samples_leaf
            snap_args['max_features'] = self.max_features
            snap_args['bootstrap'] = self.bootstrap
            snap_args['n_jobs']= self.n_jobs
            snap_args['random_state'] = self.random_state
            snap_args['use_gpu'] = self.use_gpu                          # Place Holder for now.
            snap_args['hist_type'] = self.hist_type
            snap_args['hist_nbins'] = self.hist_nbins
        
            from pai4sk import RandomForestClassifier as _snaprf
            self._snaprf = _snaprf(**snap_args)
            warnings.warn("pai4sk: RandomForestClassifier using 'snapml'", UserWarning)

        if unsupported_params:
            warnings.warn("pai4sk:unsupported parameters/value for 'snapml' : " + ", ".join([str(x) for x in unsupported_params]), UserWarning )
            warnMsg = "pai4sk: fallback to sklearn implementation due to unsupported parameters/values: " + ", ".join([str(x) for x in unsupported_params])
            warnings.warn(warnMsg, UserWarning)
            return False

        if mod_params:
            warnMsg = "pai4sk: Default values for these parameters are modified for 'snapml' : " + ", ".join([str(x) for x in mod_params])
            warnings.warn(warnMsg, UserWarning)


        return True

    def fit(self, X, y, sample_weight=None):

        """Build a forest of trees from the training set (X, y).
        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The training input samples. Internally, its dtype will be converted
            to ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csc_matrix``.
        y : array-like, shape = [n_samples] or [n_samples, n_outputs]
            The target values (class labels in classification, real numbers in
            regression).
        sample_weight : array-like, shape = [n_samples] or None
            Sample weights. If None, then samples are equally weighted. Splits
            that would create child nodes with net zero or negative weight are
            ignored while searching for a split in each node. In the case of
            classification, splits are also ignored if they would result in any
            single class carrying a negative weight in either child node.
        Returns
        -------
        self : object
        """
        if self._check_set_snapml(X, y):
            self._snaprf.fit(X,y)
            # populating attribute for snap-ml  
            self.n_outputs_ = 1
            self.classes_ = self._snaprf.uniq_labs_
            self.n_classes_ = self._snaprf.num_train_unique_labs_
            self.n_features_ = self._snaprf.n_features_
            # estimators_ , oob_score_, oob_decision_function_ are not supported for snap-ml
            self.estimators_ = None
            self.oob_score_ = 0
            self.oob_decision_function_ = [] 
        else: 
            self._randomforestsklearn.fit(X,y,sample_weight)

        return self

    def predict(self, X, num_threads=0):
        
        """Predict class for X.
        The predicted class of an input sample is a vote by the trees in
        the forest, weighted by their probability estimates. That is,
        the predicted class is the one with highest mean probability
        estimate across the trees.
        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.
        Returns
        -------
        y : array of shape = [n_samples] or [n_samples, n_outputs]
            The predicted classes.
        """
        if _check_snapml_type(X) and not (self._snaprf is None):
            return self._snaprf.predict(X, num_threads)
        else:
            return self._randomforestsklearn.predict(X)

    def predict_proba(self, X, num_threads=0):
        """Predict class probabilities for X.
        The predicted class probabilities of an input sample are computed as
        the mean predicted class probabilities of the trees in the forest. The
        class probability of a single tree is the fraction of samples of the same
        class in a leaf.
        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.

        Returns
        -------
        p : array of shape = [n_samples, n_classes], or a list of n_outputs
            such arrays if n_outputs > 1.
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute `classes_`.
            It is not supported for snap-ml
        """
        if _check_snapml_type(X) and not (self._snaprf is None):
            return self._snaprf.predict_proba(X, num_threads)
        else:
            return self._randomforestsklearn.predict_proba(X)

 
    def predict_log_proba(self, X):
        """Predict class log-probabilities for X.
        The predicted class log-probabilities of an input sample is computed as
        the log of the mean predicted class probabilities of the trees in the
        forest.
        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.
        Returns
        -------
        p : array of shape = [n_samples, n_classes], or a list of n_outputs
            such arrays if n_outputs > 1.
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute `classes_`.
        """
        if _check_snapml_type(X) and not (self._snaprf is None):
           return self._snaprf.predict_log_proba(X, num_threads)
        else:
           return self._randomforestsklearn.predict_log_proba(X)

    def apply(self, X):
        """Apply trees in the forest to X, return leaf indices.
        Parameters
        ----------
        X : array-like or sparse matrix, shape = [n_samples, n_features]
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.
        Returns
        -------
        X_leaves : array_like, shape = [n_samples, n_estimators]
            For each datapoint x in X and for each tree in the forest,
            return the index of the leaf x ends up in.
        """
        if _check_snapml_type(X) and not (self._snaprf is None):
           raise ValueError("apply() is not supported for snap-ml ; use RandomForestClassifierSklearn otherwise") 
           return np.zeros([self.n_features_,self.n_features_], dtype=np.float64)
        else:
           return self._randomforestsklearn.apply(X)

    def decision_path(self, X):
        """Return the decision path in the forest
        .. versionadded:: 0.18
        Parameters
        ----------
        X : array-like or sparse matrix, shape = [n_samples, n_features]
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.
        Returns
        -------
        indicator : sparse csr array, shape = [n_samples, n_nodes]
            Return a node indicator matrix where non zero elements
            indicates that the samples goes through the nodes.
        n_nodes_ptr : array of size (n_estimators + 1, )
            The columns from indicator[n_nodes_ptr[i]:n_nodes_ptr[i+1]]
            gives the indicator value for the i-th estimator.
        """
        if _check_snapml_type(X) and not (self._snaprf is None):
           raise ValueError("decision_path() is not supported for snap-ml ; use RandomForestClassifierSklearn otherwise") 
           return csr_matrix((1, 1), dtype=np.int8),np.array([0]).cumsum()
        else:
           return self._randomforestsklearn.decision_path(X)
 
