# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

import time
import numpy as np
import sys
import argparse
import pandas as pd

import cudf
from cudf.dataframe import DataFrame

defaultPath = "."

CLI=argparse.ArgumentParser()
CLI.add_argument(
    "--skip_sklearn",
    action = 'store_true',
    default=False
)
CLI.add_argument(
    "--data_path",
    type=str,
    default=defaultPath
)
CLI.add_argument(
   "--use_balanced_class_weights",
   action = 'store_true',
   default=False
)
CLI.add_argument(
   "--fit_intercept",
   action = 'store_true',
   default=False
)

args = CLI.parse_args()
t0 = time.time()

# Load data
pdf_trainX  = pd.read_csv(args.data_path + "/data/epsilon.X_train.csv", dtype=np.float32, header=None)
pdf_testX  = pd.read_csv(args.data_path + "/data/epsilon.X_test.csv", dtype=np.float32, header=None)
pdf_trainY = pd.read_csv(args.data_path + "/data/epsilon.y_train.csv", dtype=np.float32, header=None)
pdf_testY  = pd.read_csv(args.data_path + "/data/epsilon.y_test.csv", dtype=np.float32, header=None);

print("Data load time (s):  {0:.2f}".format(time.time()-t0))

# Convert pandas to cudf dataframes
cudf_trainX = DataFrame.from_pandas(pdf_trainX)
cudf_trainY = DataFrame.from_pandas(pdf_trainY)

# ndarray data used for training (for sklearn case)
# Converting to C-contiguous ndarray
X_train_ndarray = np.ascontiguousarray(pdf_trainX.values)
y_train_ndarray = np.ascontiguousarray(pdf_trainY.values)

# data used for training
# Create a C-contiguous DeviceNDArray from cuDF
from pai4sk.sml_io import copy_as_gpu_cmatrix
X_train = copy_as_gpu_cmatrix(cudf_trainX)
y_train = copy_as_gpu_cmatrix(cudf_trainY)

#data used for inferencing
# Converting to C-contiguous ndarray
X_test = np.ascontiguousarray(pdf_testX.values)
y_test = np.ascontiguousarray(pdf_testY.values)


num_threads = 256
cpu_gpu = "GPU"

if args.use_balanced_class_weights:
  class_weight = "balanced"
else:
  class_weight = None
alpha=1.0

# Create a LassoRegression instance from pai4sk
from pai4sk import LinearRegression
Lasso= LinearRegression(use_gpu=True, penalty="l1", num_threads=num_threads,
                        fit_intercept=args.fit_intercept, regularizer=alpha, dual=False)
# Training
t0 = time.time()
Lasso.fit(X_train, y_train)
print("[pai4sk] {0} Training time (s):  {1:.2f}".format(cpu_gpu, time.time()-t0))


# Evaluate log-loss on test set
pred = Lasso.predict(X_test)

from sklearn.metrics import mean_squared_error
mse = mean_squared_error(y_test, pred)
print("[pai4sk] Mean Squared Error :   {1:.4f}".format(cpu_gpu, mse))

#This is another way of of import 

# Import LassoRegression from pai4sk.linear_model 
from pai4sk.linear_model import Lasso
Lasso = Lasso(use_gpu=True, num_threads=num_threads,
              fit_intercept=args.fit_intercept, alpha=alpha, tol=0.001)

# Training time
t0 = time.time()
Lasso.fit(X_train, y_train)
print("[pai4sk.lmodel] Training time (s):  {0:.2f}".format(time.time()-t0))

# Inference
pred = Lasso.predict(X_test)

from pai4sk.metrics import mean_squared_error
mse = mean_squared_error(y_test, pred)
print("[pai4sk.lmodel] Mean Squared Error :   {1:.4f}".format(cpu_gpu, mse))

if args.skip_sklearn:
    print('Execution using scikit-learn is skipped.')
    exit()

# Import sklearn's LassoRegression from pai4sk module directly
from pai4sk.linear_model import LassoSklearn as Lasso
Lasso = Lasso(fit_intercept=args.fit_intercept,tol=0.001, 
              random_state=42,alpha=alpha)

# Training time
t0 = time.time()
Lasso.fit(X_train_ndarray, y_train_ndarray)
print("[sklearn] Training time (s):  {0:.2f}".format(time.time()-t0))

pred = Lasso.predict(X_test)

from pai4sk.metrics import mean_squared_error
mse = mean_squared_error(y_test, pred)
print("[sklearn] Mean Squared Error :   {1:.4f}".format(cpu_gpu, mse))
